/** Classes needed for the parsing.

  <p> Moved to antlrrt.utils:
*/
module antlrrt.parsing;

private import antlrrt.utils;
private import antlrrt.lexing;
private import antlrrt.exceptions;
private import antlrrt.collections;


// import java.lang.reflect.Constructor;


/** A circular buffer object used by TokenBuffer */
alias Queue!(Token) TokenQueue;


/**A Stream of Token objects fed to the parser from a Tokenizer that can
 * be rewound via mark()/rewind() methods.
 * <p>
 * A dynamic array is used to buffer up all the input tokens.  Normally,
 * "k" tokens are stored in the buffer.  More tokens may be stored during
 * guess mode (testing syntactic predicate), or when LT(i>k) is referenced.
 * Consumption of tokens is deferred.  In other words, reading the next
 * token is not done by conume(), but deferred until needed by LA or LT.
 * <p>
 *
 * @see antlr.Token
 * @see antlr.Tokenizer
 * @see antlr.TokenQueue
 */

public class TokenBuffer {

    // Token source
    protected TokenStream input;

    // Number of active markers
    int nMarkers = 0;

    // Additional offset used when markers are active
    int markerOffset = 0;

    // Number of calls to consume() since last LA() or LT() call
    int numToConsume = 0;

    // Circular queue
    TokenQueue queue;

    /** Create a token buffer */
    public this(TokenStream input_) {
        input = input_;
        queue = new TokenQueue(1);
    }

    /** Reset the input buffer to empty state */
    public final void reset() {
        nMarkers = 0;
        markerOffset = 0;
        numToConsume = 0;
        queue.reset();
    }

    /** Mark another token for deferred consumption */
    public final void consume() {
        numToConsume++;
    }

    /** Ensure that the token buffer is sufficiently full */
    private final void fill(int amount) {    // throws TokenStreamException 
        syncConsume();
        // Fill the buffer sufficiently to hold needed tokens
        while (queue.nbrEntries < amount + markerOffset) {
            // Append the next token
            queue.append(input.nextToken());
        }
    }

    /** return the Tokenizer (needed by ParseView) */
    public TokenStream getInput() {
        return input;
    }

    /** Get a lookahead token value */
    public final int LA(int i) {    // throws TokenStreamException
        fill(i);
        return queue.elementAt(markerOffset + i - 1).getType();
    }

    /** Get a lookahead token */
    public final Token LT(int i) {    // throws TokenStreamException
        fill(i);
        return queue.elementAt(markerOffset + i - 1);
    }

    /**Return an integer marker that can be used to rewind the buffer to
     * its current state.
     */
    public final int mark() {
        syncConsume();
//System.outp.println("Marking at " + markerOffset);
//try { for (int i = 1; i <= 2; i++) { System.outp.println("LA("+i+")=="+LT(i).getText()); } } catch (ScannerException e) {}
        nMarkers++;
        return markerOffset;
    }

    /**Rewind the token buffer to a marker.
     * @param mark Marker returned previously from mark()
     */
    public final void rewind(int mark) {
        syncConsume();
        markerOffset = mark;
        nMarkers--;
//System.outp.println("Rewinding to " + mark);
//try { for (int i = 1; i <= 2; i++) { System.outp.println("LA("+i+")=="+LT(i).getText()); } } catch (ScannerException e) {}
    }

    /** Sync up deferred consumption */
    private final void syncConsume() {
        while (numToConsume > 0) {
            if (nMarkers > 0) {
                // guess mode -- leave leading tokens and bump offset.
                markerOffset++;
            }
            else {
                // normal mode -- remove first token
                queue.removeFirst();
            }
            numToConsume--;
        }
    }
}



/** This object contains the data associated with an
 *  input stream of tokens.  Multiple parsers
 *  share a single ParserSharedInputState to parse
 *  the same stream of tokens.
 */
public class ParserSharedInputState {
    /** Where to get token objects */
    protected TokenBuffer input;

    /** Are we guessing (guessing>0)? */
    public int guessing = 0;

    /** What file (if known) caused the problem? */
    protected WString filename;

    public void reset() {
         guessing = 0;
         filename = null;
         input.reset();
     }
}



/**A generic ANTLR parser (LL(k) for k>=1) containing a bunch of
 * utility routines useful at any lookahead depth.  We distinguish between
 * the LL(1) and LL(k) parsers because of efficiency.  This may not be
 * necessary in the near future.
 *
 * Each parser object contains the state of the parse including a lookahead
 * cache (the form of which is determined by the subclass), whether or
 * not the parser is in guess mode, where tokens come from, etc...
 *
 * <p>
 * During <b>guess</b> mode, the current lookahead token(s) and token type(s)
 * cache must be saved because the token stream may not have been informed
 * to save the token (via <tt>mark</tt>) before the <tt>try</tt> block.
 * Guessing is started by:
 * <ol>
 * <li>saving the lookahead cache.
 * <li>marking the current position in the TokenBuffer.
 * <li>increasing the guessing level.
 * </ol>
 *
 * After guessing, the parser state is restored by:
 * <ol>
 * <li>restoring the lookahead cache.
 * <li>rewinding the TokenBuffer.
 * <li>decreasing the guessing level.
 * </ol>
 *
 * @see antlr.Token
 * @see antlr.TokenBuffer
 * @see antlr.LLkParser
 */

 
 
/*   TODO: Listener support
import antlr.debug.MessageListener;
import antlr.debug.ParserListener;
import antlr.debug.ParserMatchListener;
import antlr.debug.ParserTokenListener;
import antlr.debug.SemanticPredicateListener;
import antlr.debug.SyntacticPredicateListener;
import antlr.debug.TraceListener;
*/



/** 
  Factory function for AST construction
*/
typedef AST function() ASTFactoryFunc;


public abstract class Parser {
    protected ParserSharedInputState inputState;

    /** Nesting level of registered handlers */
    // protected int exceptionLevel = 0;

    /** Table of token type to token names */
    protected WString[] tokenNames;

    /** AST return value for a rule is squirreled away here */
    protected AST returnAST;

    /** AST support code; parser delegates to this object.
     *  This is set during parser construction by default
     *  to either "new ASTFactory()" or a ctor that
     *  has a token type to class map for hetero nodes.
     */
    protected ASTFactory astFactory = null;

    /** Constructed if any AST types specified in tokens{..}.
     *  Maps an Integer->Class object.
     */
    protected Hashtable!(int, ASTFactoryFunc) tokenTypeToASTClassMap = null;

    private boolean ignoreInvalidDebugCalls = false;

    /** Used to keep track of indentdepth for traceIn/Out */
    protected int traceDepth = 0;

    public this() {
        this(new ParserSharedInputState());
    }

    public this(ParserSharedInputState state) {
        inputState = state;
    }

    /** If the user specifies a tokens{} section with heterogeneous
     *  AST node types, then ANTLR generates code to fill
     *  this mapping.
     */
    public Hashtable!(int, ASTFactoryFunc) getTokenTypeToASTClassMap() {
        return tokenTypeToASTClassMap;
    }

/*
    public void addMessageListener(MessageListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addMessageListener() is only valid if parser built for debugging");
    }

    public void addParserListener(ParserListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addParserListener() is only valid if parser built for debugging");
    }

    public void addParserMatchListener(ParserMatchListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addParserMatchListener() is only valid if parser built for debugging");
    }

    public void addParserTokenListener(ParserTokenListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addParserTokenListener() is only valid if parser built for debugging");
    }

    public void addSemanticPredicateListener(SemanticPredicateListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addSemanticPredicateListener() is only valid if parser built for debugging");
    }

    public void addSyntacticPredicateListener(SyntacticPredicateListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addSyntacticPredicateListener() is only valid if parser built for debugging");
    }

    public void addTraceListener(TraceListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("addTraceListener() is only valid if parser built for debugging");
    }
*/

    /**Get another token object from the token stream */
    public abstract void consume();    // throws TokenStreamException;

    /** Consume tokens until one matches the given token */
    public void consumeUntil(int tokenType) {    // throws TokenStreamException
        while (LA(1) != Token.EOF_TYPE && LA(1) != tokenType) {
            consume();
        }
    }

    /** Consume tokens until one matches the given token set */
    public void consumeUntil(BitSet set) {    // throws TokenStreamException
        while (LA(1) != Token.EOF_TYPE && !set.member(LA(1))) {
            consume();
        }
    }

    protected void defaultDebuggingSetup(TokenStream lexer, TokenBuffer tokBuf) {
        // by default, do nothing -- we're not debugging
    }

    /** Get the AST return value squirreled away in the parser */
    public AST getAST() {
        return returnAST;
    }

    public ASTFactory getASTFactory() {
        return astFactory;
    }

    public WString getFilename() {
        return inputState.filename;
    }

    public ParserSharedInputState getInputState() {
        return inputState;
    }

    public void setInputState(ParserSharedInputState state) {
        inputState = state;
    }

    public WString getTokenName(int num) {
        return tokenNames[num];
    }

    public WString[] getTokenNames() {
        return tokenNames;
    }

    public boolean isDebugMode() {
        return false;
    }

    /** Return the token type of the ith token of lookahead where i=1
     * is the current token being examined by the parser (i.e., it
     * has not been matched yet).
     */
    public abstract int LA(int i);    // throws TokenStreamException;

    /**Return the ith token of lookahead */
    public abstract Token LT(int i);    // throws TokenStreamException;

    // Forwarded to TokenBuffer
    public int mark() {
        return inputState.input.mark();
    }

    /**Make sure current lookahead symbol matches token type <tt>t</tt>.
     * Throw an exception upon mismatch, which is catch by either the
     * error handler or by the syntactic predicate.
     */
    public void match(int t) {   // throws MismatchedTokenException, TokenStreamException
        if (LA(1) != t)
            throw new MismatchedTokenException(tokenNames, LT(1), t, false, getFilename());
        else
        // mark token as consumed -- fetch next token deferred until LA/LT
            consume();
    }

    /**Make sure current lookahead symbol matches the given set
     * Throw an exception upon mismatch, which is catch by either the
     * error handler or by the syntactic predicate.
     */
    public void match(BitSet b) {   // throws MismatchedTokenException, TokenStreamException
        if (!b.member(LA(1)))
            throw new MismatchedTokenException(tokenNames, LT(1), b, false, getFilename());
        else
        // mark token as consumed -- fetch next token deferred until LA/LT
            consume();
    }

    public void matchNot(int t) {   // throws MismatchedTokenException, TokenStreamException
        if (LA(1) == t)
        // Throws inverted-sense exception
            throw new MismatchedTokenException(tokenNames, LT(1), t, true, getFilename());
        else
        // mark token as consumed -- fetch next token deferred until LA/LT
            consume();
    }

    /** @deprecated as of 2.7.2. This method calls System.exit() and writes
     *  directly to stderr, which is usually not appropriate when
     *  a parser is embedded into a larger application. Since the method is
     *  <code>static</code>, it cannot be overridden to avoid these problems.
     *  ANTLR no longer uses this method internally or in generated code.
     */
    public static void panic() {
        System.err.println(asWString("Parser: panic"));
        System.exit(1);
    }

/*
    public void removeMessageListener(MessageListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new RuntimeException("removeMessageListener() is only valid if parser built for debugging");
    }

    public void removeParserListener(ParserListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new RuntimeException("removeParserListener() is only valid if parser built for debugging");
    }

    public void removeParserMatchListener(ParserMatchListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new RuntimeException("removeParserMatchListener() is only valid if parser built for debugging");
    }

    public void removeParserTokenListener(ParserTokenListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new RuntimeException("removeParserTokenListener() is only valid if parser built for debugging");
    }

    public void removeSemanticPredicateListener(SemanticPredicateListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("removeSemanticPredicateListener() is only valid if parser built for debugging");
    }

    public void removeSyntacticPredicateListener(SyntacticPredicateListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new IllegalArgumentException("removeSyntacticPredicateListener() is only valid if parser built for debugging");
    }

    public void removeTraceListener(TraceListener l) {
        if (!ignoreInvalidDebugCalls)
            throw new RuntimeException("removeTraceListener() is only valid if parser built for debugging");
    }
*/

    /** Parser error-reporting function can be overridden in subclass */
    public void reportError(RecognitionException ex) {
        System.err.println(ex.getMessage());
    }

    /** Parser error-reporting function can be overridden in subclass */
    public void reportError(WString s) {
        if (getFilename() is null) {
            System.err.println("error: " ~ s);
        }
        else {
            System.err.println(getFilename() ~ ": error: " ~ s);
        }
    }

    /** Parser warning-reporting function can be overridden in subclass */
    public void reportWarning(WString s) {
        if (getFilename() is null) {
            System.err.println("warning: " ~ s);
        }
        else {
            System.err.println(getFilename() ~ ": warning: " ~ s);
        }
    }

    public void recover(RecognitionException ex,
                        BitSet tokenSet) {    // throws TokenStreamException
        consume();
        consumeUntil(tokenSet);
    }

    public void rewind(int pos) {
        inputState.input.rewind(pos);
    }

    /** Specify an object with support code (shared by
     *  Parser and TreeParser.  Normally, the programmer
     *  does not play with this, using setASTNodeType instead.
     */
    public void setASTFactory(ASTFactory f) {
        astFactory = f;
    }

    public void setASTNodeClass(WString cl) {
        astFactory.setASTNodeType(cl);
    }

    /** Specify the type of node to create during tree building; use setASTNodeClass now
     *  to be consistent with Token Object Type accessor.
     *  @deprecated since 2.7.1
     */
    public void setASTNodeType(WString nodeType) {
        setASTNodeClass(nodeType);
    }

    public void setDebugMode(boolean debugMode) {
        if (!ignoreInvalidDebugCalls)
            throw new RuntimeException(asWString("setDebugMode() only valid if parser built for debugging"));
    }

    public void setFilename(WString f) {
        inputState.filename = f;
    }

    public void setIgnoreInvalidDebugCalls(boolean value) {
        ignoreInvalidDebugCalls = value;
    }

    /** Set or change the input token buffer */
    public void setTokenBuffer(TokenBuffer t) {
        inputState.input = t;
    }

    public void traceIndent() {
        for (int i = 0; i < traceDepth; i++)
            System.outp.print(" ");
    }

    public void traceIn(WString rname) {    // throws TokenStreamException
        traceDepth += 1;
        traceIndent();
        System.outp.println("> " ~ rname ~ "; LA(1)==" ~ LT(1).getText() ~
                           asWString((inputState.guessing > 0)?" [guessing]":""));
    }

    public void traceOut(WString rname) {    // throws TokenStreamException
        traceIndent();
        System.outp.println("< " ~ rname ~ "; LA(1)==" ~ LT(1).getText() ~
                           asWString((inputState.guessing > 0)?" [guessing]":""));
        traceDepth -= 1;
    }
}



/**An LL(k) parser.
 *
 * @see antlr.Token
 * @see antlr.TokenBuffer
 */
public class LLkParser : Parser {
    int k;

    public this(int k_) {
        k = k_;
    }

    public this(ParserSharedInputState state, int k_) {
        super(state);
        k = k_;
    }

    public this(TokenBuffer tokenBuf, int k_) {
        k = k_;
        setTokenBuffer(tokenBuf);
    }

    public this(TokenStream lexer, int k_) {
        k = k_;
        TokenBuffer tokenBuf = new TokenBuffer(lexer);
        setTokenBuffer(tokenBuf);
    }

    /**Consume another token from the input stream.  Can only write sequentially!
     * If you need 3 tokens ahead, you must consume() 3 times.
     * <p>
     * Note that it is possible to overwrite tokens that have not been matched.
     * For example, calling consume() 3 times when k=2, means that the first token
     * consumed will be overwritten with the 3rd.
     */
    public void consume() {    // throws TokenStreamException
        inputState.input.consume();
    }

    public int LA(int i) {    // throws TokenStreamException
        return inputState.input.LA(i);
    }

    public Token LT(int i) {    // throws TokenStreamException
        return inputState.input.LT(i);
    }

    private void trace(WString ee, WString rname) {    // throws TokenStreamException
        traceIndent();
        System.outp.print(ee ~ rname ~ asWString((inputState.guessing > 0)?"; [guessing]":"; "));
        for (int i = 1; i <= k; i++) {
            if (i != 1) {
                System.outp.print(", ");
            }
            if ( LT(i) !is null ) {
                System.outp.print("LA(" ~ asWString(i) ~ ")==" ~ LT(i).getText());
            }
            else {
                System.outp.print("LA(" ~ asWString(i) ~ ") is null");
            }
        }
        System.outp.println("");
    }

    public void traceIn(WString rname) {    // throws TokenStreamException
        traceDepth += 1;
        trace(asWString("> "), rname);
    }

    public void traceOut(WString rname) {    // throws TokenStreamException
        trace(asWString("< "), rname);
        traceDepth -= 1;
    }
}



/** This object contains the data associated with an
 *  input AST.  Multiple parsers
 *  share a single TreeParserSharedInputState to parse
 *  the same tree or to have the parser walk multiple
 *  trees.
 */
public class TreeParserSharedInputState {
    /** Are we guessing (guessing>0)? */
    public int guessing = 0;
}


public class TreeParser {
    /** The AST Null object; the parsing cursor is set to this when
     *  it is found to be null.  This way, we can test the
     *  token type of a node without having to have tests for null
     *  everywhere.
     */
    public static ASTNULLType ASTNULL;  // = new ASTNULLType();
    static this()
    {
        ASTNULL = new ASTNULLType();
    }

    /** Where did this rule leave off parsing; avoids a return parameter */
    protected AST _retTree;

    /** guessing nesting level; guessing==0 implies not guessing */
    // protected int guessing = 0;

    /** Nesting level of registered handlers */
    // protected int exceptionLevel = 0;

    protected TreeParserSharedInputState inputState;

    /** Table of token type to token names */
    protected WString[] tokenNames;

    /** AST return value for a rule is squirreled away here */
    protected AST returnAST;

    /** AST support code; parser and treeparser delegate to this object */
    protected ASTFactory astFactory;  //  = new ASTFactory();

    /** Used to keep track of indentdepth for traceIn/Out */
    protected int traceDepth = 0;

    public this() {
        astFactory = new ASTFactory();
        inputState = new TreeParserSharedInputState();
    }

    /** Get the AST return value squirreled away in the parser */
    public AST getAST() {
        return returnAST;
    }

    public ASTFactory getASTFactory() {
        return astFactory;
    }

    public WString getTokenName(int num) {
        return tokenNames[num];
    }

    public WString[] getTokenNames() {
        return tokenNames;
    }

    protected void match(AST t, int ttype) {    // throws MismatchedTokenException
        //System.outp.println("match("+ttype+"); cursor is "+t);
        if (t is null || t is ASTNULL || t.getType() != ttype) {
            throw new MismatchedTokenException(getTokenNames(), t, ttype, false);
        }
    }

    /**Make sure current lookahead symbol matches the given set
     * Throw an exception upon mismatch, which is catch by either the
     * error handler or by the syntactic predicate.
     */
    public void match(AST t, BitSet b) {    // throws MismatchedTokenException
        if (t is null || t is ASTNULL || !b.member(t.getType())) {
            throw new MismatchedTokenException(getTokenNames(), t, b, false);
        }
    }

    protected void matchNot(AST t, int ttype) {    // throws MismatchedTokenException
        //System.outp.println("match("+ttype+"); cursor is "+t);
        if (t is null || t is ASTNULL || t.getType() == ttype) {
            throw new MismatchedTokenException(getTokenNames(), t, ttype, true);
        }
    }

    /** @deprecated as of 2.7.2. This method calls System.exit() and writes
     *  directly to stderr, which is usually not appropriate when
     *  a parser is embedded into a larger application. Since the method is 
     *  <code>static</code>, it cannot be overridden to avoid these problems.
     *  ANTLR no longer uses this method internally or in generated code.
     */
    public static void panic() {
        System.err.println("TreeWalker: panic");
        System.exit(1);
    }

    /** Parser error-reporting function can be overridden in subclass */
    public void reportError(RecognitionException ex) {
        System.err.println(ex.toWString());
    }

    /** Parser error-reporting function can be overridden in subclass */
    public void reportError(WString s) {
        System.err.println("error: " ~ s);
    }

    /** Parser warning-reporting function can be overridden in subclass */
    public void reportWarning(WString s) {
        System.err.println("warning: " ~ s);
    }

    /** Specify an object with support code (shared by
     *  Parser and TreeParser.  Normally, the programmer
     *  does not play with this, using setASTNodeType instead.
     */
    public void setASTFactory(ASTFactory f) {
        astFactory = f;
    }

    /** Specify the type of node to create during tree building.
     *     @deprecated since 2.7.2
     */
    public void setASTNodeType(WString nodeType) {
        setASTNodeClass(nodeType);
    }

    /** Specify the type of node to create during tree building */
    public void setASTNodeClass(WString nodeType) {
        astFactory.setASTNodeType(nodeType);
    }

    public void traceIndent() {
        for (int i = 0; i < traceDepth; i++)
            System.outp.print(" ");
    }

    public void traceIn(WString rname, AST t) {
        traceDepth += 1;
        traceIndent();
        System.outp.println("> " ~ rname ~
                           "(" ~ (t !is null?t.toWString():asWString("null")) ~ ")" ~
                           asWString((inputState.guessing > 0)?" [guessing]":""));
    }

    public void traceOut(WString rname, AST t) {
        traceIndent();
        System.outp.println("< " ~ rname ~
                           "(" ~ (t !is null?t.toWString():asWString("null")) ~ ")" ~
                           asWString((inputState.guessing > 0)?" [guessing]":""));
        traceDepth--;
    }
}


public interface ASTEnumeration {
    public boolean hasMoreNodes();

    public AST nextNode();
}


public class ASTEnumerator : /*implements*/ ASTEnumeration {
    /** The list of root nodes for subtrees that match */
    // VectorEnumerator nodes;
    Vector!(AST) vec;
    int i = 0;

    public this(Vector!(AST) v) {
        //nodes = new VectorEnumerator(v);
        vec = v;
    }

    public boolean hasMoreNodes() {
        synchronized (vec) {
            return i < vec.length(); //nodes.vector.lastElement;
        }
    }

    public AST nextNode() {
        synchronized (vec) {
            if (hasMoreNodes()) {
                return vec.get(i++);    // (AST)nodes.vector.data[i++];
            }
            throw new NoSuchElementException(asWString("ASTEnumerator"));
        }
    }
}



/** Minimal AST node interface used by ANTLR AST generation
 * and tree-walker.
 */
public interface AST {
    /// Needed for D for XML serialization
    public WString className();

    /** Add a (rightmost) child to this node */
    public void addChild(AST c);

    public boolean equals(AST t);

    public boolean equalsList(AST t);

    public boolean equalsListPartial(AST t);

    public boolean equalsTree(AST t);

    public boolean equalsTreePartial(AST t);

    public ASTEnumeration findAll(AST tree);

    public ASTEnumeration findAllPartial(AST subtree);

    /** Get the first child of this node; null if no children */
    public AST getFirstChild();

    /** Get    the next sibling in line after this one */
    public AST getNextSibling();

    /** Get the token text for this node */
    public WString getText();

    /** Get the token type for this node */
    public int getType();

    /** @since 2.7.3 Need for error handling */
    public int getLine();

    /** @since 2.7.3 Need for error handling */
    public int getColumn();

    /** Get number of children of this node; if leaf, returns 0 */
    public int getNumberOfChildren();

    public void initialize(int t, WString txt);

    public void initialize(AST t);

    public void initialize(Token t);

    /** Set the first child of a node. */
    public void setFirstChild(AST c);

    /** Set the next sibling after this one. */
    public void setNextSibling(AST n);

    /** Set the token text for this node */
    public void setText(WString text);

    /** Set the token type for this node */
    public void setType(int ttype);

    public WString toWString();

    public WString toStringList();

    public WString toStringTree();
}


/** There is only one instance of this class **/
public class ASTNULLType : /*implements*/ AST {
    public WString className()
    {
        return asWString("antlrrt.ASTNULLType");
    }
    
    public void addChild(AST c) {
    }

    public boolean equals(AST t) {
        return false;
    }

    public boolean equalsList(AST t) {
        return false;
    }

    public boolean equalsListPartial(AST t) {
        return false;
    }

    public boolean equalsTree(AST t) {
        return false;
    }

    public boolean equalsTreePartial(AST t) {
        return false;
    }

    public ASTEnumeration findAll(AST tree) {
        return null;
    }

    public ASTEnumeration findAllPartial(AST subtree) {
        return null;
    }

    public AST getFirstChild() {
        return this;
    }

    public AST getNextSibling() {
        return this;
    }

    public WString getText() {
        return asWString("<ASTNULL>");
    }

    public int getType() {
        return Token.NULL_TREE_LOOKAHEAD;
    }

    public int getLine() {
        return 0;
    }

    public int getColumn() {
        return 0;
    }

    public int getNumberOfChildren() {
        return 0;
    }

    public void initialize(int t, WString txt) {
    }

    public void initialize(AST t) {
    }

    public void initialize(Token t) {
    }

    public void setFirstChild(AST c) {
    }

    public void setNextSibling(AST n) {
    }

    public void setText(WString text) {
    }

    public void setType(int ttype) {
    }

    public WString toWString() {
        return getText();
    }

    public WString toStringList() {
        return getText();
    }

    public WString toStringTree() {
        return getText();
    }
}



/**
 * A Child-Sibling Tree.
 *
 * A tree with PLUS at the root and with two children 3 and 4 is
 * structured as:
 *
 *        PLUS
 *          |
 *          3 -- 4
 *
 * and can be specified easily in LISP notation as
 *
 * (PLUS 3 4)
 *
 * where every '(' starts a new subtree.
 *
 * These trees are particular useful for translators because of
 * the flexibility of the children lists.  They are also very easy
 * to walk automatically, whereas trees with specific children
 * reference fields can't easily be walked automatically.
 *
 * This class contains the basic support for an AST.
 * Most people will create ASTs that are subclasses of
 * BaseAST or of CommonAST.
 */
public abstract class BaseAST : /*implements*/ AST {   // implements Serializable
    protected BaseAST down;
    protected BaseAST right;

    private static boolean verboseStringConversion = false;
    private static WString[] tokenNames = null;
    
    
    public WString className()
    {
        return asWString(this.classinfo.name);
    }

    /**Add a node to the end of the child list for this node */
    public void addChild(AST node) {
        if (node is null) return;
        BaseAST t = this.down;
        if (t !is null) {
            while (t.right !is null) {
                t = t.right;
            }
            t.right = cast(BaseAST)node;
        }
        else {
            this.down = cast(BaseAST)node;
        }
    }

    /** How many children does this node have? */
    public int getNumberOfChildren() {
        BaseAST t = this.down;
        int n = 0;
        if (t !is null) {
            n = 1;
            while (t.right !is null) {
                t = t.right;
                n++;
            }
            return n;
        }
        return n;
    }

    // Original: private void doWorkForFindAll(Vector v, AST target, boolean partialMatch) {
    private void doWorkForFindAll(Vector!(AST) v, AST target, boolean partialMatch) {
        AST sibling;

        // Start walking sibling lists, looking for matches.
        siblingWalk:
        for (sibling = this;
             sibling !is null;
             sibling = sibling.getNextSibling()) {
            if ((partialMatch && sibling.equalsTreePartial(target)) ||
                (!partialMatch && sibling.equalsTree(target))) {
                v.appendElement(sibling);
            }
            // regardless of match or not, check any children for matches
            if (sibling.getFirstChild() !is null) {
                (cast(BaseAST)sibling.getFirstChild()).doWorkForFindAll(v, target, partialMatch);
            }
        }
    }

    /** Is node t equal to this in terms of token type and text? */
    public boolean equals(AST t) {
        if (t is null) return false;
        if ( (this.getText() is null && t.getText() !is null) ||
             (this.getText() !is null && t.getText() is null) )
        {
            return false;
        }
        if ( this.getText() is null && t.getText() is null ) {
            return this.getType() == t.getType();
        }
        return this.getText() == t.getText() &&
            this.getType() == t.getType();
    }

    /** Is t an exact structural and equals() match of this tree.  The
     *  'this' reference is considered the start of a sibling list.
     */
    public boolean equalsList(AST t) {
        AST sibling;

        // the empty tree is not a match of any non-null tree.
        if (t is null) {
            return false;
        }

        // Otherwise, start walking sibling lists.  First mismatch, return false.
        for (sibling = this;
             sibling !is null && t !is null;
             sibling = sibling.getNextSibling(), t = t.getNextSibling())
        {
            // as a quick optimization, check roots first.
            if (!sibling.equals(t)) {
                return false;
            }
            // if roots match, do full list match test on children.
            if (sibling.getFirstChild() !is null) {
                if (!sibling.getFirstChild().equalsList(t.getFirstChild())) {
                    return false;
                }
            }
            // sibling has no kids, make sure t doesn't either
            else if (t.getFirstChild() !is null) {
                return false;
            }
        }
        if (sibling is null && t is null) {
            return true;
        }
        // one sibling list has more than the other
        return false;
    }

    /** Is 'sub' a subtree of this list?
     *  The siblings of the root are NOT ignored.
     */
    public boolean equalsListPartial(AST sub) {
        AST sibling;

        // the empty tree is always a subset of any tree.
        if (sub is null) {
            return true;
        }

        // Otherwise, start walking sibling lists.  First mismatch, return false.
        for (sibling = this;
             sibling !is null && sub !is null;
             sibling = sibling.getNextSibling(), sub = sub.getNextSibling()) {
            // as a quick optimization, check roots first.
            if (!sibling.equals(sub)) return false;
            // if roots match, do partial list match test on children.
            if (sibling.getFirstChild() !is null) {
                if (!sibling.getFirstChild().equalsListPartial(sub.getFirstChild())) return false;
            }
        }
        if (sibling is null && sub !is null) {
            // nothing left to match in this tree, but subtree has more
            return false;
        }
        // either both are null or sibling has more, but subtree doesn't
        return true;
    }

    /** Is tree rooted at 'this' equal to 't'?  The siblings
     *  of 'this' are ignored.
     */
    public boolean equalsTree(AST t) {
        // check roots first.
        if (!this.equals(t)) return false;
        // if roots match, do full list match test on children.
        if (this.getFirstChild() !is null) {
            if (!this.getFirstChild().equalsList(t.getFirstChild())) return false;
        }
        // sibling has no kids, make sure t doesn't either
        else if (t.getFirstChild() !is null) {
            return false;
        }
        return true;
    }

    /** Is 't' a subtree of the tree rooted at 'this'?  The siblings
     *  of 'this' are ignored.
     */
    public boolean equalsTreePartial(AST sub) {
        // the empty tree is always a subset of any tree.
        if (sub is null) {
            return true;
        }

        // check roots first.
        if (!this.equals(sub)) return false;
        // if roots match, do full list partial match test on children.
        if (this.getFirstChild() !is null) {
            if (!this.getFirstChild().equalsListPartial(sub.getFirstChild())) return false;
        }
        return true;
    }

    /** Walk the tree looking for all exact subtree matches.  Return
     *  an ASTEnumerator that lets the caller walk the list
     *  of subtree roots found herein.
     */
    public ASTEnumeration findAll(AST target) {
        Vector!(AST) roots = new Vector!(AST)(10);
        AST sibling;

        // the empty tree cannot result in an enumeration
        if (target is null) {
            return null;
        }

        doWorkForFindAll(roots, target, false);  // find all matches recursively

        return new ASTEnumerator(roots);
    }

    /** Walk the tree looking for all subtrees.  Return
     *  an ASTEnumerator that lets the caller walk the list
     *  of subtree roots found herein.
     */
    public ASTEnumeration findAllPartial(AST sub) {
        Vector!(AST) roots = new Vector!(AST)(10);
        AST sibling;

        // the empty tree cannot result in an enumeration
        if (sub is null) {
            return null;
        }

        doWorkForFindAll(roots, sub, true);  // find all matches recursively

        return new ASTEnumerator(roots);
    }

    /** Get the first child of this node; null if not children */
    public AST getFirstChild() {
        return down;
    }

    /** Get the next sibling in line after this one */
    public AST getNextSibling() {
        return right;
    }

    /** Get the token text for this node */
    public WString getText() {
        return asWString("");
    }

    /** Get the token type for this node */
    public int getType() {
        return 0;
    }

    public int getLine() {
        return 0;
    }

    public int getColumn() {
        return 0;
    }

    public abstract void initialize(int t, WString txt);

    public abstract void initialize(AST t);

    public abstract void initialize(Token t);

    /** Remove all children */
    public void removeChildren() {
        down = null;
    }

    public void setFirstChild(AST c) {
        down = cast(BaseAST)c;
    }

    public void setNextSibling(AST n) {
        right = cast(BaseAST)n;
    }

    /** Set the token text for this node */
    public void setText(WString text) {
    }

    /** Set the token type for this node */
    public void setType(int ttype) {
    }

    public static void setVerboseStringConversion(boolean verbose, WString[] names) {
        verboseStringConversion = verbose;
        tokenNames = names;
    }

    /** Return an array of strings that maps token ID to it's text. @since 2.7.3 */
    public static WString[] getTokenNames() {
        return tokenNames;
    }

    public WString toWString() {
        StringBuffer b = new StringBuffer();
        // if verbose and type name not same as text (keyword probably)
        if (verboseStringConversion &&
            !getText().equalsIgnoreCase(tokenNames[getType()]) &&
            !getText().equalsIgnoreCase(StringUtils.stripFrontBack(tokenNames[getType()], asWString("\""), asWString("\"")))) {
            b.append('[');
            b.append(getText());
            b.append(",<");
            b.append(tokenNames[getType()]);
            b.append(">]");
            return b.toWString();
        }
        return getText();
    }

    /** Print out a child-sibling tree in LISP notation */
    public WString toStringList() {
        AST t = this;
        WString ts = asWString("");
        
        if (t.getFirstChild() !is null)
            appS(ts, " (");
        appS(ts, " " ~ this.toWString());
        if (t.getFirstChild() !is null) {
            appS(ts, (cast(BaseAST)t.getFirstChild()).toStringList());
        }
        if (t.getFirstChild() !is null)
            appS(ts, " )");
        if (t.getNextSibling() !is null) {
            appS(ts, (cast(BaseAST)t.getNextSibling()).toStringList());
        }
        return ts;
    }

    public WString toStringTree() {
        AST t = this;
        WString ts = asWString("");
        
        if (t.getFirstChild() !is null)
            appS(ts, " (");
        appS(ts, " " ~ this.toWString());
        if (t.getFirstChild() !is null) {
            appS(ts, (cast(BaseAST)t.getFirstChild()).toStringList());
        }
        if (t.getFirstChild() !is null)
            appS(ts, " )");
        return ts;
    }

    public static WString decode(WString text) {
        wchar c, c1, c2, c3, c4, c5;
        StringBuffer n = new StringBuffer();
        for (int i = 0; i < text.length; i++) {
            c = text.charAt(i);
            if (c == '&') {
                c1 = text.charAt(i + 1);
                c2 = text.charAt(i + 2);
                c3 = text.charAt(i + 3);
                c4 = text.charAt(i + 4);
                c5 = text.charAt(i + 5);

                if (c1 == 'a' && c2 == 'm' && c3 == 'p' && c4 == ';') {
                    n.append("&");
                    i += 5;
                }
                else if (c1 == 'l' && c2 == 't' && c3 == ';') {
                    n.append("<");
                    i += 4;
                }
                else if (c1 == 'g' && c2 == 't' && c3 == ';') {
                    n.append(">");
                    i += 4;
                }
                else if (c1 == 'q' && c2 == 'u' && c3 == 'o' &&
                    c4 == 't' && c5 == ';') {
                    n.append("\"");
                    i += 6;
                }
                else if (c1 == 'a' && c2 == 'p' && c3 == 'o' &&
                    c4 == 's' && c5 == ';') {
                    n.append("'");
                    i += 6;
                }
                else
                    n.append("&");
            }
            else
                n.append(c);
        }
        return n.toWString(); //new String(n);
    }

    public static WString encode(WString text) {
        wchar c;
        StringBuffer n = new StringBuffer();
        for (int i = 0; i < text.length; i++) {
            c = text.charAt(i);
            switch (c) {
                case '&':
                    {
                        n.append("&amp;");
                        break;
                    }
                case '<':
                    {
                        n.append("&lt;");
                        break;
                    }
                case '>':
                    {
                        n.append("&gt;");
                        break;
                    }
                case '"':
                    {
                        n.append("&quot;");
                        break;
                    }
                case '\'':
                    {
                        n.append("&apos;");
                        break;
                    }
                default :
                    {
                        n.append(c);
                        break;
                    }
            }
        }
        return n.toWString(); //new String(n);
    }

    public void xmlSerializeNode(Writer outp)
        {   // throws IOException 
        StringBuffer buf = new StringBuffer(100);
        buf.append("<");
        buf.append(className() ~ " ");
        buf.append("text=\"" ~ encode(getText()) ~ "\" type=\"" ~
                   asWString(getType()) ~ "\"/>");
        outp.write(buf.toWString());
    }

    public void xmlSerializeRootOpen(Writer outp)
        {   // throws IOException 
        StringBuffer buf = new StringBuffer(100);
        buf.append("<");
        buf.append(className() ~ " ");
        buf.append("text=\"" ~ encode(getText()) ~ "\" type=\"" ~
                   asWString(getType()) ~ "\">\n");
        outp.write(buf.toWString());
    }

    public void xmlSerializeRootClose(Writer outp)
        {   // throws IOException 
        outp.write("</" ~ className() ~ ">\n");
    }

    public void xmlSerialize(Writer outp) {  // throws IOException 
        // print out this node and all siblings
        for (AST node = this;
             node !is null;
             node = node.getNextSibling()) {
            if (node.getFirstChild() is null) {
                // print guts (class name, attributes)
                (cast(BaseAST)node).xmlSerializeNode(outp);
            }
            else {
                (cast(BaseAST)node).xmlSerializeRootOpen(outp);

                // print children
                (cast(BaseAST)node.getFirstChild()).xmlSerialize(outp);

                // print end tag
                (cast(BaseAST)node).xmlSerializeRootClose(outp);
            }
        }
    }
}


/** Common AST node implementation */
public class CommonAST : BaseAST {
    int ttype = Token.INVALID_TYPE;
    WString text;
    
    // static const WString CLASS_NAME = 
    
    WString className()
    {
        return asWString("antlrrt.CommonAST");
    }
    
    public static AST antlrFactory()
    {
        return new CommonAST();
    }

    /** Get the token text for this node */
    public WString getText() {
        return text;
    }

    /** Get the token type for this node */
    public int getType() {
        return ttype;
    }

    public void initialize(int t, WString txt) {
        setType(t);
        setText(txt);
    }

    public void initialize(AST t) {
        setText(t.getText());
        setType(t.getType());
    }

    public this() {
    }

    public this(Token tok) {
        initialize(tok);
    }

    public void initialize(Token tok) {
        setText(tok.getText());
        setType(tok.getType());
    }

    /** Set the token text for this node */
    public void setText(WString text_) {
        text = text_;
    }

    /** Set the token type for this node */
    public void setType(int ttype_) {
        ttype = ttype_;
    }
}



/** A CommonAST whose initialization copies hidden token
 *  information from the Token used to create a node.
 */
public class CommonASTWithHiddenTokens : CommonAST {
    protected CommonHiddenStreamToken hiddenBefore, hiddenAfter; // references to hidden tokens

    WString className()
    {
        return asWString("antlrrt.CommonASTWithHiddenTokens");
    }
    
    public static AST antlrFactory()
    {
        return new CommonASTWithHiddenTokens();
    }
    
    public this() {
        super();
    }

    public this(Token tok) {
        super(tok);
    }

    public CommonHiddenStreamToken getHiddenAfter() {
        return hiddenAfter;
    }

    public CommonHiddenStreamToken getHiddenBefore() {
        return hiddenBefore;
    }

    public void initialize(AST t)
    {
        hiddenBefore = (cast(CommonASTWithHiddenTokens)t).getHiddenBefore();
        hiddenAfter = (cast(CommonASTWithHiddenTokens)t).getHiddenAfter();
        super.initialize(t);
    }

    public void initialize(Token tok) {
        CommonHiddenStreamToken t = cast(CommonHiddenStreamToken)tok;
        super.initialize(t);
        hiddenBefore = t.getHiddenBefore();
        hiddenAfter = t.getHiddenAfter();
    }
}



/** ASTPair:  utility class used for manipulating a pair of ASTs
 * representing the current AST root and current AST sibling.
 * This exists to compensate for the lack of pointers or 'var'
 * arguments in Java.
 */
public class ASTPair {
    public AST root;        // current root of tree
    public AST child;        // current child to which siblings are added

    /** Make sure that child is the last sibling */
    public final void advanceChildToEnd() {
        if (child !is null) {
            while (child.getNextSibling() !is null) {
                child = child.getNextSibling();
            }
        }
    }

    /** Copy an ASTPair.  Don't call it clone() because we want type-safety */
    public ASTPair copy() {
        ASTPair tmp = new ASTPair();
        tmp.root = root;
        tmp.child = child;
        return tmp;
    }

    public WString toWString() {
        WString r = root is null ? asWString("null") : root.getText();
        WString c = child is null ? asWString("null") : child.getText();
        return "[" ~ r ~ "," ~ c ~ "]";
    }
}


/** ASTArray is a class that allows ANTLR to
 * generate code that can create and initialize an array
 * in one expression, like:
 *    (new ASTArray(3)).add(x).add(y).add(z)
 */
public class ASTArray {
    public int size = 0;
    public AST[] array;


    public this(int capacity) {
        array = new AST[capacity];
    }

    public ASTArray add(AST node) {
        array[size++] = node;
        return this;
    }
}


/** AST Support code shared by TreeParser and Parser.
 *  We use delegation to share code (and have only one
 *  bit of code to maintain) rather than subclassing
 *  or superclassing (forces AST support code to be
 *  loaded even when you don't want to do AST stuff).
 *
 *  <p>Typically, setASTNodeType is used to specify the
 *  homogeneous type of node to create, but you can override
 *  create to make heterogeneous nodes etc...
 
    <p>Because D hasn't such nice reflection features as Java,
    the D version handles pointers to factory function instead
    of Class objects, even if the names of the member variables
    are not changed and contain the word "class".
 */
 
public class ASTFactory {
    /** Name of AST class to create during tree construction.
     *  Null implies that the create method should create
     *  a default AST type such as CommonAST.  This is for
     *  homogeneous nodes.
     */
    protected WString theASTNodeType = null;
    protected ASTFactoryFunc theASTNodeTypeClass = null;
    
    /** Needed in D to map a classname to its ASTFactoryFunc.
    */
    protected ASTFactoryFunc[WString] classMap;
    
    public void registerClass(WString className, ASTFactoryFunc cls)
    {
        classMap[className] = cls;
    }
    
    public void registerClass(wchar[] className, ASTFactoryFunc cls)
    {
        classMap[new WString(className)] = cls;
    }
    
    
    /** Replaces Class.forName()
    */
    public ASTFactoryFunc classForName(WString className)
    {
        ASTFactoryFunc* f = className in classMap;
        
        if (!f)
            throw new IllegalArgumentException("Invalid class, " ~ className);
        
        return *f;
    }
    
    /** How to specify the classname to create for a particular
     *  token type.  Note that ANTLR allows you to say, for example,
     *
        tokens {
         PLUS<AST=PLUSNode>;
         ...
        }
     *
     *  and it tracks everything statically.  #[PLUS] will make you
     *  a PLUSNode w/o use of this table.
     *
     *  For tokens that ANTLR cannot track statically like #[i],
     *  you can use this table to map PLUS (Integer) -> PLUSNode (Class)
     *  etc... ANTLR sets the class map from the tokens {...} section
     *  via the ASTFactory(Hashtable) ctor in antlr.Parser.
     */
    protected Hashtable!(int, ASTFactoryFunc) tokenTypeToASTClassMap = null;

    public this() {
        registerClass(asWString("antlrrt.CommonAST"), &CommonAST.antlrFactory);
        registerClass(asWString("antlrrt.CommonASTWithHiddenTokens"), &CommonASTWithHiddenTokens.antlrFactory);
    }

    /** Create factory with a specific mapping from token type
     *  to Java AST node type.  Your subclasses of ASTFactory
     *  can override and reuse the map stuff.
     */
    public this(Hashtable!(int, ASTFactoryFunc) tokenTypeToClassMap) {
        this();
        setTokenTypeToASTClassMap(tokenTypeToClassMap);
    }
    
    /** Specify an "override" for the Java AST object created for a
     *  specific token.  It is provided as a convenience so
     *  you can specify node types dynamically.  ANTLR sets
     *  the token type mapping automatically from the tokens{...}
     *  section, but you can change that mapping with this method.
     *  ANTLR does it's best to statically determine the node
     *  type for generating parsers, but it cannot deal with
     *  dynamic values like #[LT(1)].  In this case, it relies
     *  on the mapping.  Beware differences in the tokens{...}
     *  section and what you set via this method.  Make sure
     *  they are the same.
     *
     *  Set className to null to remove the mapping.
     *
     *  @since 2.7.2
     */
    public void setTokenTypeASTNodeType(int tokenType, WString className)
        // throws IllegalArgumentException
    {
        if ( tokenTypeToASTClassMap is null ) {
            tokenTypeToASTClassMap = new Hashtable!(int, ASTFactoryFunc)();
        }
        if ( className is null ) {
            tokenTypeToASTClassMap.remove(tokenType);
            return;
        }
        ASTFactoryFunc c = null;
        try {
            c = classForName(className);
            tokenTypeToASTClassMap.put(tokenType, c);
        }
        catch (Exception e) {
            throw new IllegalArgumentException("Invalid class, " ~ className);
        }
    }
    
    /** Convenience method to set token type and register the AST class
      at once.
    */
    public void setTokenTypeASTNodeType(int tokenType, WString className,
            ASTFactoryFunc cls)
        // throws IllegalArgumentException
    {
        registerClass(className, cls);
        setTokenTypeASTNodeType(tokenType, className);
    }
    

    /** For a given token type, what is the AST node object type to create
     *  for it?
     *  @since 2.7.2
     */
    public ASTFactoryFunc getASTNodeType(int tokenType) {
        // try node specific class
        if ( tokenTypeToASTClassMap !is null ) {
            ASTFactoryFunc c = tokenTypeToASTClassMap.get(tokenType);
            if ( c !is null ) {
                return c;
            }
        }

        // try a global specified class
        if (theASTNodeTypeClass !is null) {
            return theASTNodeTypeClass;
        }

        // default to the common type
        return &CommonAST.antlrFactory;
    }

    /** Add a child to the current AST */
    public void addASTChild(ASTPair currentAST, AST child) {
        if (child !is null) {
            if (currentAST.root is null) {
                // Make new child the current root
                currentAST.root = child;
            }
            else {
                if (currentAST.child is null) {
                    // Add new child to current root
                    currentAST.root.setFirstChild(child);
                }
                else {
                    currentAST.child.setNextSibling(child);
                }
            }
            // Make new child the current child
            currentAST.child = child;
            currentAST.advanceChildToEnd();
        }
    }

    /** Create a new empty AST node; if the user did not specify
     *  an AST node type, then create a default one: CommonAST.
     */
    public AST create() {
        return create(Token.INVALID_TYPE);
    }

    public AST create(int type) {
        ASTFactoryFunc c = getASTNodeType(type);
        AST t = create(c);
        if ( t !is null ) {
            t.initialize(type, asWString(""));
        }
        return t;
    }

    public AST create(int type, WString txt) {
        AST t = create(type);
        if ( t !is null ) {
            t.initialize(type, txt);
        }
        return t;
    }

    /** Create an AST node with the token type and text passed in, but
     *  with a specific Java object type. Typically called when you
     *  say @[PLUS,"+",PLUSNode] in an antlr action.
     *  @since 2.7.2
     */
    public AST create(int type, WString txt, WString className) {
        return create(type, txt, classForName(className));
    }

        
    public AST create(int type, WString txt, ASTFactoryFunc cls) {
        AST t = create(cls);
        if ( t !is null ) {
            t.initialize(type, txt);
        }
        return t;
    }


    /** Create a new empty AST node; if the user did not specify
     *  an AST node type, then create a default one: CommonAST.
     */
    public AST create(AST tr) {
        if (tr is null) return null;        // create(null) is null
        AST t = create(tr.getType());
        if ( t !is null ) {
            t.initialize(tr);
        }
        return t;
    }

    public AST create(Token tok) {
        AST t = create(tok.getType());
        if ( t !is null ) {
            t.initialize(tok);
        }
        return t;
    }

    /** ANTLR generates reference to this when you reference a token
     *  that has a specified heterogeneous AST node type.  This is
     *  also a special case node creation routine for backward
     *  compatibility.  Before, ANTLR generated "new T(tokenObject)"
     *  and so I must call the appropriate constructor not T().
     *
     * @since 2.7.2
     */
    public AST create(Token tok, WString className) {
        return createUsingCtor(tok, classForName(className));
    }
    
    public AST create(Token tok, ASTFactoryFunc cls) {
        return createUsingCtor(tok, cls);
    }

    /**
     * @since 2.7.2
     */
    public AST create(WString className) {
        return create(classForName(className));
    }

    /**
     * @since 2.7.2
     */   // TODO ?!
    protected AST createUsingCtor(Token token, WString className) {
        return createUsingCtor(token, classForName(className));
    }
    
    
    protected AST createUsingCtor(Token token, ASTFactoryFunc cls) {
        AST t = null;
        try {
            /*
            Class[] tokenArgType = new Class[] { antlr.Token.class };
            try {
                Constructor ctor = c.getConstructor(tokenArgType);
                t = (AST)ctor.newInstance(new Object[]{token}); // make a new one
            }
            catch (NoSuchMethodException e){
            */
                // just do the regular thing if you can't find the ctor
                // Your AST must have default ctor to use this.
                t = create(cls);
                if ( t !is null ) {
                    t.initialize(token);
                }
            //}
        }
        catch (Exception e) {
            throw new IllegalArgumentException(asWString("Invalid class or can't make instance"));  // ~ className);
        }
        return t;
    }

    /**
     * @since 2.7.2
     */
    protected AST create(ASTFactoryFunc cls) {
        AST t = null;
        try {
            t = cls(); // make a new one
        }
        catch (Exception e) {
            error(asWString("Can't create AST Node "));  // + c.getName());
            return null;
        }
        return t;
    }

    /** Copy a single node with same Java AST objec type.
     *  Ignore the tokenType->Class mapping since you know
     *  the type of the node, t.getClass(), and doing a dup.
     *
     *  clone() is not used because we want all AST creation
     *  to go thru the factory so creation can be
     *  tracked.  Returns null if t is null.
     */
    public AST dup(AST t) {
        if ( t is null ) {
            return null;
        }
        AST dup_t = create(t.className());    // getClass());
        dup_t.initialize(t);
        return dup_t;
    }

    /** Duplicate tree including siblings of root. */
    public AST dupList(AST t) {
        AST result = dupTree(t);            // if t is null, then result is null
        AST nt = result;
        while (t !is null) {                        // for each sibling of the root
            t = t.getNextSibling();
            nt.setNextSibling(dupTree(t));    // dup each subtree, building new tree
            nt = nt.getNextSibling();
        }
        return result;
    }

    /**Duplicate a tree, assuming this is a root node of a tree--
     * duplicate that node and what's below; ignore siblings of root node.
     */
    public AST dupTree(AST t) {
        AST result = dup(t);        // make copy of root
        // copy all children of root.
        if (t !is null) {
            result.setFirstChild(dupList(t.getFirstChild()));
        }
        return result;
    }

    /** Make a tree from a list of nodes.  The first element in the
     *  array is the root.  If the root is null, then the tree is
     *  a simple list not a tree.  Handles null children nodes correctly.
     *  For example, build(a, b, null, c) yields tree (a b c).  build(null,a,b)
     *  yields tree (nil a b).
     */
    public AST make(AST[] nodes) {
        if (nodes is null || nodes.length == 0) return null;
        AST root = nodes[0];
        AST tail = null;
        if (root !is null) {
            root.setFirstChild(null);    // don't leave any old pointers set
        }
        // link in children;
        for (int i = 1; i < nodes.length; i++) {
            if (nodes[i] is null) continue;    // ignore null nodes
            if (root is null) {
                // Set the root and set it up for a flat list
                root = tail = nodes[i];
            }
            else if (tail is null) {
                root.setFirstChild(nodes[i]);
                tail = root.getFirstChild();
            }
            else {
                tail.setNextSibling(nodes[i]);
                tail = tail.getNextSibling();
            }
            // Chase tail to last sibling
            while (tail.getNextSibling() !is null) {
                tail = tail.getNextSibling();
            }
        }
        return root;
    }

    /** Make a tree from a list of nodes, where the nodes are contained
     * in an ASTArray object
     */
    public AST make(ASTArray nodes) {
        return make(nodes.array);
    }

    /** Make an AST the root of current AST */
    public void makeASTRoot(ASTPair currentAST, AST root) {
        if (root !is null) {
            // Add the current root as a child of new root
            root.addChild(currentAST.root);
            // The new current child is the last sibling of the old root
            currentAST.child = currentAST.root;
            currentAST.advanceChildToEnd();
            // Set the new root
            currentAST.root = root;
        }
    }

//  Original:    public void setASTNodeClass(Class c) {   // TODO ?
    public void setASTNodeClass(ASTFactoryFunc c, WString t) {
        if ( c !is null ) {
            theASTNodeTypeClass = c;
            theASTNodeType = t;   //c.getName();
        }
    }

    public void setASTNodeClass(WString t) {
        theASTNodeType = t;
        try {
            theASTNodeTypeClass = classForName(t); // get class def
        }
        catch (Exception e) {
            // either class not found,
            // class is interface/abstract, or
            // class or initializer is not accessible.
            error("Can't find/access AST Node type" ~ t);
        }
    }

    /** Specify the type of node to create during tree building.
     *     @deprecated since 2.7.1
     */
    public void setASTNodeType(WString t) {
        setASTNodeClass(t);
    }

    public Hashtable!(int, ASTFactoryFunc) getTokenTypeToASTClassMap() {
        return tokenTypeToASTClassMap;
    }

    public void setTokenTypeToASTClassMap(Hashtable!(int, ASTFactoryFunc) tokenTypeToClassMap) {
        this.tokenTypeToASTClassMap = tokenTypeToClassMap;
    }

    /** To change where error messages go, can subclass/override this method
     *  and then setASTFactory in Parser and TreeParser.  This method removes
     *  a prior dependency on class antlr.Tool.
     */
    public void error(WString e) {
        System.err.println(e);
    }
}


public interface ASTVisitor {
    public void visit(AST node);
}
